#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (C) 2010 Tobi Vollebregt

from ConfigParser import RawConfigParser as ConfigParser
from progressbar import ProgressBar
import rapid
import getopt, os, sys


def usage():
	print """Usage: %(progname)s <verb> [<argument>]

Where verb is one of:
 * update|upgrade: Install the latest package for all pinned tags.
 * pin: Pins a tag and installs the latest package for that tag.
 * unpin: Unpins a tag. Does not uninstall any packages.
 * install: Install a package. Does not pin any tags.
 * uninstall: Uninstall a package. Unpin its tag if any.
 * list-tags: List all tags that contain <argument>.
 * list-pinned-tags: Idem, but only pinned tags.
 * list-packages: List all packages whose name contains <argument>.
 * list-installed-packages: Idem, but only installed packages.

Examples:
%(progname)s pin xta:latest           # installs latest XTA
%(progname)s pin 's44:latest mutator' # installs latest Spring: 1944
%(progname)s upgrade                  # upgrade all pinned tags
""" % {'progname': sys.argv[0]}
	sys.exit(1)


# Read configuration (i.e. list of pinned tags)
config_path = os.path.join(rapid.content_dir, 'main.cfg')
config = ConfigParser()
config.read(config_path)
pinned_tags = []

if not config.has_section('tags'):
	config.add_section('tags')
if config.has_option('tags', 'pinned'):
	pinned_tags = set([s for s in config.get('tags', 'pinned').split(',') if s])


#  Create rapid module.
rapid = rapid.Rapid()


def select(noun, needle, haystack):
	n = needle.lower()
	selected = filter(lambda s: n in str(s).lower(), haystack)

	if len(selected) == 0:
		print 'No %ss matching %s found.' % (noun, needle)
		sys.exit(1)

	if len(selected) >= 100:
		print '100 or more matching %ss found, please narrow your search.' % noun
		sys.exit(1)

	if len(selected) > 1:
		print 'Multiple %ss found:' % noun
		for i in range(len(selected)):
			print '%2i.  %s' % (i + 1, selected[i])
		which = raw_input("Which %s do you mean? (enter number or 'all')   " % noun)
		if which == 'all':
			return selected
		try:
			which = int(which) - 1
		except ValueError:
			sys.exit(1)
		return [selected[which]]

	return selected


def pin_single(tag):
	""" Pin a tag. This means any package having this tag will automatically be
	    installed and upgraded."""
	if not tag in pinned_tags:
		print 'Pinning: ' + tag
		pinned_tags.add(tag)
	else:
		print 'Already pinned: ' + tag


def pin(searchterm):
	""" Pin all tags matching searchterm and install the corresponding packages."""
	for t in select('tag', searchterm, rapid.get_packages_by_tag().iterkeys()):
		pin_single(t)
		install_single(rapid.get_package_by_tag(t))


def unpin_single(tag):
	""" Unpin a tag. This means packages having this tag will not be
	    automatically upgraded anymore. Does not uninstall anything."""
	if tag in pinned_tags:
		print 'Unpinning: ' + tag
		pinned_tags.remove(tag)
	else:
		print 'Not pinned: ' + tag


def unpin(searchterm):
	""" Unpin all tags matching searchterm."""
	for t in select('pinned tag', searchterm, pinned_tags):
		unpin_single(t)


def install_single(p, dep = False):
	""" Install a single package and its dependencies."""
	if p:
		for d in p.dependencies:
			install_single(d, True)
		if not p.installed():
			print ['Installing: ', 'Installing dependency: '][int(dep)] + p.name
			p.install(ProgressBar())
			print
		elif not dep:
			print 'Already installed: ' + p.name


def install(searchterm):
	""" Install all packages matching searchterm."""
	for name in select('name', searchterm, rapid.get_names().iterkeys()):
		install_single(rapid.get_package_by_name(name))


def uninstall_single(p):
	""" Uninstall and unpin a single package. Does not uninstall dependencies."""
	if p:
		if not p.can_be_uninstalled():
			print 'Can not uninstall because of dependencies: ' + p.name
			return
		for t in p.tags:
			unpin_single(t)
		print 'Uninstalling: ' + p.name
		p.uninstall()


def uninstall(searchterm):
	""" Uninstall all packages matching searchterm."""
	for name in select('name', searchterm, [p.name for p in rapid.get_installed_packages()]):
		uninstall_single(rapid.get_package_by_name(name))


def list_packages(searchterm, available):
	""" List all packages whose name matches searchterm."""
	s = searchterm.lower()
	print 'Installed packages:'
	for p in filter(lambda p: s in p.name.lower(), rapid.get_installed_packages()):
		print '  %-40s (%s)' % (p.name, ', '.join(p.tags))
	if available:
		print 'Available packages:'
		for p in rapid.get_not_installed_packages():
			print '  %-40s (%s)' % (p.name, ', '.join(p.tags))


def list_tags(searchterm, available):
	""" List all tags which match searchterm."""
	s = searchterm.lower()
	print 'Pinned tags:'
	for tag in filter(lambda t: s in t.lower(), pinned_tags):
		p = rapid.get_package_by_tag(tag)
		if p:
			print '  %-40s (%s)' % (tag, p.name)
		else:
			print '  %-40s [dangling tag]' % tag
	if available:
		print 'Available tags:'
		for tag in (set(rapid.get_packages_by_tag()) - pinned_tags):
			p = rapid.get_package_by_tag(tag)
			print '  %-40s (%s)' % (tag, p.name)


def upgrade(searchterm):
	""" Upgrade installed tags which match searchterm."""
	for tag in filter(lambda t: searchterm.lower() in t.lower(), pinned_tags):
		install_single(rapid.get_package_by_tag(tag))


def req_arg():
	if len(sys.argv) < 3:
		print 'Not enough arguments to operation: ' + verb
		print
		usage()
	return sys.argv[2]

def opt_arg():
	if len(sys.argv) > 2:
		return sys.argv[2]
	return ''



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (C) 2010 Tobi Vollebregt

from contextlib import closing
import ConfigParser
import os
import urllib2

################################################################################

timeout = 5     # change timeout here if desired

################################################################################

def atomic_write(filename, data):
	temp = filename + '.tmp'
	with open(temp, 'wb') as f:
		f.write(data)
	if os.path.exists(filename): # on Windows rename doesn't overwrite destination
		os.remove(filename)
	os.rename(temp, filename)


class NotModifiedHandler(urllib2.BaseHandler):

	def http_error_304(self, req, fp, code, message, headers):
		addinfourl = urllib2.addinfourl(fp, headers, req.get_full_url())
		addinfourl.code = code
		return addinfourl

################################################################################

class Downloader:
	__config = ConfigParser.RawConfigParser()
	_304 = False    # for unit tests

	def __init__(self, config_filename):
		#print ('reading configuration from ' + config_filename)
		self.__config_filename = config_filename
		self.__config.read(config_filename)

	def __write_config(self):
		#print ('writing configuration to ' + self.__config_filename)
		with open(self.__config_filename, 'wb') as f:
			self.__config.write(f)

	def __config_get(self, section, option):
		if self.__config.has_option(section, option):
			return self.__config.get(section, option)

	def __config_set(self, section, option, value):
		if value:
			if not self.__config.has_section(section):
				self.__config.add_section(section)
			self.__config.set(section, option, value)

	def onetime_get_request(self, url, filename):
		if os.path.exists(filename):
			return

		request = urllib2.Request(url, unverifiable=True)

		with closing(urllib2.urlopen(request, timeout = timeout)) as remote:
			atomic_write(filename, remote.read())

	def conditional_get_request(self, url, filename):
		section = url + ',' + filename
		etag = self.__config_get(section, 'etag')
		last_modified = self.__config_get(section, 'last_modified')

		request = urllib2.Request(url, unverifiable=True)

		if os.path.exists(filename):
			if etag:
				request.add_header('If-None-Match', etag)
			if last_modified:
				request.add_header('If-Modified-Since', last_modified)

		try:
			with closing(urllib2.build_opener(NotModifiedHandler()).open(request, timeout = timeout)) as remote:
				headers = remote.info()
				self.__config_set(section, 'etag', headers.getheader('ETag'))
				self.__config_set(section, 'last_modified', headers.getheader('Last-Modified'))
				self.__write_config()

				if hasattr(remote, 'code') and remote.code == 304:
					#print 'the file has not been modified'
					self._304 = True
					return

				atomic_write(filename, remote.read())

		except urllib2.URLError:
			if os.path.exists(filename):
				return
			raise

	def post(self, url, data):
		request = urllib2.Request(url, unverifiable=True)
		request.add_data(data)

		return urllib2.urlopen(request)

################################################################################

from StringIO import StringIO

class MockDownloader:
	def __init__(self):
		self.www = {}
		self.visited = set()
		self._304 = False

	def onetime_get_request(self, url, filename):
		atomic_write(filename, self.www[url])

	def conditional_get_request(self, url, filename):
		if url in self.visited:
			self._304 = True
			return

		atomic_write(filename, self.www[url])
		self.visited.add(url)

	def post(self, url, data):
		def info():
			return {'Content-Length': len(self.www[url])}
		remote = StringIO(self.www[url])
		remote.info = info
		return remote

################################################################################

import unittest, shutil

class TestDownloader(unittest.TestCase):
	test_dir = '.test-downloader'
	url = 'http://repos.caspring.org/repos.gz'
	test_file = os.path.join(test_dir, 'repos.gz')
	config_file = os.path.join(test_dir, 'test.cfg')

	def setUp(self):
		os.mkdir(self.test_dir)
		if False:   # change to test (Mock)Downloader
			self.downloader = Downloader(self.config_file)
		else:
			self.downloader = MockDownloader()
			self.downloader.www[self.url] = ''

	def tearDown(self):
		shutil.rmtree(self.test_dir)

	def test_onetime_get_request(self):
		self.downloader.onetime_get_request(self.url, self.test_file)

	def test_http_304_not_modified(self):
		self.downloader.conditional_get_request(self.url, self.test_file)
		self.assertFalse(self.downloader._304)
		self.downloader.conditional_get_request(self.url, self.test_file)
		self.assertTrue(self.downloader._304)

	def test_post(self):
		#TODO: implement test_post
		pass

if __name__ == '__main__':
	unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys

############################################################
#
# A progress bar that actually shows progress!
#
# Source:
# http://code.activestate.com/recipes/168639-progress-bar-class/
#
############################################################

class ProgressBar:
	""" Creates a text-based progress bar. Call the object with the `print'
		command to see the progress bar, which looks something like this:

		[=======>		22%				  ]

		You may specify the progress bar's width, min and max values on init.
	"""

	def __init__(self, minValue = 0, maxValue = 100, totalWidth=80):
		self.progBar = "[]"   # This holds the progress bar string
		self.min = minValue
		self.max = maxValue
		self.width = totalWidth
		self.amount = minValue           # When amount == max, we are 100% done
		self.update_amount(self.amount)  # Build progress bar string

	def update_amount(self, newAmount = 0):
		""" Update the progress bar with the new amount (with min and max
			values set at initialization; if it is over or under, it takes the
			min or max value as a default. """
		if newAmount < self.min: newAmount = self.min
		if newAmount > self.max: newAmount = self.max
		self.amount = newAmount

		# Figure out the new percent done, round to an integer
		diffFromMin = float(self.amount - self.min)
		percentDone = (diffFromMin / float(self.max - self.min)) * 100.0
		percentDone = int(round(percentDone))

		# Figure out how many hash bars the percentage should be
		allFull = self.width - 2
		numHashes = (percentDone / 100.0) * allFull
		numHashes = int(round(numHashes))

		# Build a progress bar with an arrow of equal signs; special cases for
		# empty and full
		if numHashes == 0:
			self.progBar = "[>%s]" % (' '*(allFull-1))
		elif numHashes == allFull:
			self.progBar = "[%s]" % ('='*allFull)
		else:
			self.progBar = "[%s>%s]" % ('='*(numHashes-1),
										' '*(allFull-numHashes))

		# figure out where to put the percentage, roughly centered
		percentPlace = (len(self.progBar) / 2) - len(str(percentDone))
		percentString = str(percentDone) + "%"

		# slice the percentage into the bar
		self.progBar = ''.join([self.progBar[0:percentPlace], percentString,
								self.progBar[percentPlace+len(percentString):]
								])

	def __str__(self):
		return str(self.progBar)

	def __call__(self, value):
		""" Increases the amount by value, and writes to stdout. Prints a
		    carriage return first, so it will overwrite the current line in
		    stdout."""
		print '\r',
		self.update_amount(self.amount + value)
		sys.stdout.write(str(self))
		sys.stdout.write("\r")
		sys.stdout.flush()

	def setMaximum(self, value):
		self.max = value

	def maximum(self):
		return self.max

if __name__ == '__main__':
	from time import sleep
	p = ProgressBar()
	for i in range(0, 201):
		p(1)
		if i == 90:
			p.max = 200
		sleep(0.02)
	print



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from PyQt4 import QtCore, QtGui
import rapid,main

class BaseRapidModel(QtGui.QStandardItemModel):
	def __init__(self,parent):
		super(BaseRapidModel,self).__init__(parent)

	def reloadData(self, dataFunction):
		self.clear()
		self.setColumnCount(2)
		self.setRowCount(0)
		self.setHeaderData(0, QtCore.Qt.Horizontal, "Name")
		self.setHeaderData(1, QtCore.Qt.Horizontal, "Tags")
		i = 0
		for p in filter( dataFunction, main.rapid.get_packages() ):
			self.insertRow(i)
			self.setData(self.index(i, 0), p.name)
			self.setData(self.index(i, 1), ', '.join(p.tags))
			i += 1

class AvailableRapidModel(BaseRapidModel):
	def __init__(self,parent):
		super(AvailableRapidModel, self).__init__(parent)

	def reload(self):
		self.reloadData( lambda p: not p.installed() )

class InstalledRapidModel(BaseRapidModel):
	def __init__(self,parent):
		super(InstalledRapidModel, self).__init__(parent)

	def reload(self):
		self.reloadData( lambda p: p.installed() )


# -*- coding: utf-8 -*-
from distutils.core import setup

setup(
    name='rapid-spring',
    version='0.1.1',
    author='Tobi Vollebregt',
    author_email='',
    packages=['rapid'],
    scripts=['bin/rapid','bin/rapid-gui'],
    url='http://pypi.python.org/pypi/rapid-spring/',
    license='LICENSE.txt',
    description='spring contentt downloading',
    long_description=open('README.txt').read(),
)


#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (C) 2010 Tobi Vollebregt

from bitarray import bitarray
from contextlib import closing
from hashlib import md5
from urlparse import urlparse
from StringIO import StringIO
import binascii, gzip, os, struct

from downloader import Downloader, atomic_write

################################################################################

# content_dir : Storage for temporary files (repos.gz, versions.gz)
# spring_dir  : Spring data directory
# pool_dir    : Where pool files are stored (visible to Spring)
# package_dir : Where package files are stored (visible to Spring)

master_url = 'http://repos.caspring.org/repos.gz'

def set_spring_dir(path):
	global spring_dir, pool_dir, package_dir, content_dir
	spring_dir = path
	pool_dir = os.path.join(spring_dir, 'pool')
	package_dir = os.path.join(spring_dir, 'packages')
	content_dir = os.path.join(spring_dir, 'rapid')

if os.name == 'posix':
	set_spring_dir(os.path.join(os.environ['HOME'], '.spring'))
#FIXME: elif os.name =='nt':
else:
	raise NotImplementedError('Unknown OS: %s' % os.name)

################################################################################

def mkdir(path):
	""" Create directory if it does not exist yet. """
	if not os.path.exists(path):
		os.mkdir(path)

def mkdir_p(path):
	""" Create directories if they do not exist yet. """
	if not os.path.exists(path):
		os.makedirs(path)

def psv(s):
	""" Split pipe separated value string into list of non-empty components."""
	return [x for x in s.split('|') if x]

def gzip_string(s):
	""" Gzip the string s."""
	fileobj = StringIO()
	with closing(gzip.GzipFile(mode = 'wb', fileobj = fileobj)) as f:
		f.write(s)
	return fileobj.getvalue()

################################################################################

class RapidException(Exception):
	""" Base class for other exceptions defined in this module."""
	pass


class PackageFormatException(RapidException):
	""" Raised when a .sdp package can not be read."""
	def __init__(self, field):
		self.field = field

	def __str__(self):
		return self.field


class StreamerFormatException(RapidException):
	""" Raised when the output of streamer.cgi can not be read."""
	def __init__(self, field):
		self.field = field

	def __str__(self):
		return self.field

################################################################################

class Rapid:
	""" Repository container."""
	def __init__(self):
		self.__repositories = None
		self.__packages = None
		self.__packages_by_tag = None
		self.master_url = master_url
		self.cache_dir = content_dir
		self.repos_gz = os.path.join(self.cache_dir, 'repos.gz')
		self.packages_gz = os.path.join(self.cache_dir, 'packages.gz')

		mkdir(spring_dir)
		mkdir(content_dir)
		mkdir(package_dir)

		if not os.path.exists(pool_dir):
			os.mkdir(pool_dir)
			for i in range(0, 256):
				os.mkdir(os.path.join(pool_dir, '%02x' % i))

		self.downloader = Downloader(os.path.join(content_dir, 'downloader.cfg'))

	def update(self):
		""" Update of the master list of repositories."""
		self.downloader.conditional_get_request(self.master_url, self.repos_gz)

	def get_repositories(self):
		""" Download and return list of repositories."""
		if self.__repositories:
			return self.__repositories

		self.update()
		with closing(gzip.open(self.repos_gz)) as f:
			unique = set([x.split(',')[1] for x in f])
			self.__repositories = [Repository(self, x) for x in unique]

		return self.__repositories

	def read_packages_gz(self):
		""" Reads global packages.gz into a dictionary of Packages.

			Contrary to versions.gz, packages.gz:
			- is normalised (i.e. every package occurs only once),
			- does not support '|' characters in tags (tags are '|' separated)
		"""
		def read_line(line):
			row = line[:-1].split(',')
			return (row[3], Package(tags         = psv(row[0]), hex  = row[1],
			                        dependencies = psv(row[2]), name = row[3]))

		if os.path.exists(self.packages_gz):
			with closing(gzip.open(self.packages_gz)) as f:
				return dict(map(read_line, f))
		return {}

	def write_packages_gz(self):
		# Write to temporary file
		tempfile = self.packages_gz + '.tmp'
		with closing(gzip.open(tempfile, 'wb')) as f:
			for p in self.__packages.itervalues():
				# tags, hex, dependencies, name
				f.write(','.join(['|'.join(p.tags), p.hex, '|'.join(p.dependencies), p.name]) + '\n')

		# Commit by moving temporary file over packages.gz
		if os.path.exists(tempfile):
			if os.path.exists(self.packages_gz):
				os.remove(self.packages_gz)
			os.rename(tempfile, self.packages_gz)

	def get_packages_by_name(self):
		""" Return package with given name or None if there isn't any."""
		if self.__packages:
			return self.__packages

		self.__packages = self.read_packages_gz()

		# FIXME: this is broken if a package is in repo1 with tag1 and in repo2 with tag2
		for r in self.get_repositories():
			self.__packages.update(r.get_packages())

		self.write_packages_gz()

		# Resolve dependencies and calculate reverse dependencies.
		# Dependencies missing in all repositories are silently discarded.
		for p in self.__packages.itervalues():
			p.dependencies = set([self.__packages[name] for name in p.dependencies if name in self.__packages])
			for d in p.dependencies:
				d.reverse_dependencies.add(p)

		return self.__packages

	def get_package_by_name(self, name):
		""" Return package with given name or None if there isn't any."""
		if name in self.get_packages_by_name():
			return self.__packages[name]

		return None

	def get_packages_by_tag(self):
		""" Return a dictionary mapping tag to Package."""
		if self.__packages_by_tag:
			return self.__packages_by_tag

		self.__packages_by_tag = {}
		for p in self.get_packages_by_name().itervalues():
			self.__packages_by_tag.update(dict([(t, p) for t in p.tags]))

		return self.__packages_by_tag

	def get_package_by_tag(self, tag):
		""" Return package with given tag or None if there isn't any."""
		if tag in self.get_packages_by_tag():
			return self.__packages_by_tag[tag]

		return None

	def get_packages(self):
		""" Return something that can iterate over all packages."""
		return self.get_packages_by_name().itervalues()

	def get_installed_packages(self):
		""" Return something that can iterate over all installed packages."""
		return filter(lambda p: p.installed(), self.get_packages())

	def get_not_installed_packages(self):
		""" Return something that can iterate over all not-installed packages."""
		return filter(lambda p: not p.installed(), self.get_packages())

################################################################################

class Repository:
	""" A rapid package repository."""
	def __init__(self, rapid, url):
		self.__packages = None
		self.rapid = rapid
		self.url = url
		self.cache_dir = os.path.join(self.rapid.cache_dir, urlparse(url).netloc)
		self.package_cache_dir = os.path.join(self.cache_dir, 'packages')
		self.versions_gz = os.path.join(self.cache_dir, 'versions.gz')

		mkdir(self.cache_dir)
		mkdir(self.package_cache_dir)

	def update(self):
		""" Update of the list of packages of this repository."""
		self.rapid.downloader.conditional_get_request(self.url + '/versions.gz', self.versions_gz)

	def read_versions_gz(self):
		""" Reads versions.gz-formatted file into a dictionary of Packages."""
		packages = {}

		def read_line(line):
			row = line[:-1].split(',')   # tag,hex,dependencies,name
			tag, hex, deps, name = row[0], row[1], psv(row[2]), row[3]
			if not name in packages:
				packages[name] = Package(hex, name, deps, repository = self)
			assert (packages[name].hex == hex)
			assert (packages[name].dependencies == deps)
			assert (packages[name].name == name)
			packages[name].tags.add(tag)

		with closing(gzip.open(self.versions_gz)) as f:
			map(read_line, f)

		return packages

	def get_packages(self):
		""" Download and return the list of packages offered. For these
		    packages dependencies have not been resolved to other Package
		    objects, because of cross-repository dependencies."""
		if self.__packages:
			return self.__packages

		self.update()
		self.__packages = self.read_versions_gz()
		return self.__packages

################################################################################

class Package:
	def __init__(self, hex, name, dependencies, tags = None, repository = None):
		self.__files = None
		self.hex = hex
		self.name = name
		self.dependencies = dependencies
		self.reverse_dependencies = set()
		self.tags = set(tags or [])
		self.repository = repository
		if repository:
			self.cache_file = os.path.join(repository.package_cache_dir, self.hex + '.sdp')

	def get_installed_path(self):
		""" Return the path at which the package would be visible to Spring."""
		return os.path.join(package_dir, self.hex + '.sdp')

	def download(self):
		""" Download the package from the repository."""
		self.repository.rapid.downloader.onetime_get_request(self.repository.url + '/packages/' + self.hex + '.sdp', self.cache_file)

	def get_files(self):
		""" Download .sdp file and return the list of files in it."""
		if self.__files:
			return self.__files
		self.__files = []

		self.download()
		with closing(gzip.open(self.cache_file)) as f:
			def really_read(n, field):
				data = f.read(n)
				if len(data) < n:
					raise PackageFormatException(field)
				return data

			while True:
				namelen = f.read(1)
				if namelen == '': break   # normal loop termination condition
				namelen = struct.unpack('B', namelen)[0]

				name  = really_read(namelen, 'name')
				md5   = really_read(16, 'md5')
				crc32 = really_read(4, 'crc32')
				size  = really_read(4, 'size')

				size = struct.unpack('>L', size)[0]
				self.__files.append(File(self, name, md5, crc32, size))

		return self.__files

	def download_files(self, requested_files, progress = None):
		""" Download requested_files using the streamer.cgi interface.

		    Progress is reported through the progress object, which must be
		    callable (with a single argument to indicate progress _increase_),
		    a setMaximum (int) setter and int maximum() getter

		    streamer.cgi works as follows:
		    * The client does a POST to /streamer.cgi?<hex>
		      Where hex = the name of the .sdp
		    * The client then sends a gzipped bitarray representing the files
		      it wishes to download. Bitarray is formated in the obvious way,
		      an array of characters where each file in the sdp is represented
		      by the (index mod 8) bit (shifted left) of the (index div 8) byte
		      of the array.
		    * streamer.cgi then responds with <big endian encoded int32 length>
		      <data of gzipped pool file> for all files requested. Files in the
		      pool are also gzipped, so there is no need to decompress unless
		      you wish to verify integrity.
		    * streamer.cgi also sets the Content-Length header in the reply so
		      you can implement a proper progress bar.

		"""
		# Determine which files to fetch. (as bitarray and list of files)
		requested_files = set(requested_files)
		bits = bitarray(map(lambda f: f in requested_files, self.get_files()), endian='little')
		expected_files = filter(lambda f: f in requested_files, self.get_files())
		if len(expected_files) == 0:
			return

		# Build HTTP POST data.
		postdata = gzip_string(bits.tostring())

		# Perform HTTP POST request and download and process the response.
		with closing(self.repository.rapid.downloader.post(self.repository.url + '/streamer.cgi?' + self.hex, postdata)) as remote:
			if not remote.info().has_key('Content-Length'):
				raise StreamerFormatException('Content-Length')

			if progress:
				progress.setMaximum( int(remote.info()['Content-Length']) )
				progress(0)

			for f in expected_files:
				size = remote.read(4)
				if size == '': raise StreamerFormatException('size')
				size = struct.unpack('>L', size)[0]

				data = remote.read(size)
				if len(data) < size: raise StreamerFormatException('data')

				# check md5 hash
				with closing(gzip.GzipFile(mode = 'rb', fileobj = StringIO(data))) as g:
					if md5(g.read()).digest() != f.md5:
						raise StreamerFormatException('md5')

				mkdir_p( os.path.dirname(f.get_pool_path()) )
				atomic_write(f.get_pool_path(), data)

				if progress:
					progress(4 + size)

	def get_missing_files(self):
		""" Return a list of files which are not locally available."""
		return filter(lambda f: not f.available(), self.get_files())

	def can_be_installed(self):
		""" Return true iff all dependencies are installed."""
		if not self.installed():
			for dep in self.dependencies:
				if not dep.installed():
					return False
		return True

	def install(self, progress = None):
		""" Install the package by hardlinking it into Spring dir."""
		if not self.installed() and self.can_be_installed():
			self.download_files(self.get_missing_files(), progress)
			#FIXME: Windows support
			os.link(self.cache_file, self.get_installed_path())
			if progress:
				progress(progress.maximum())

	def can_be_uninstalled(self):
		""" Return true iff no reverse dependencies are installed."""
		if self.installed():
			for rdep in self.reverse_dependencies:
				if rdep.installed():
					return False
		return True

	def uninstall(self):
		""" Uninstall the package by unlinking it from Spring dir."""
		if self.installed() and self.can_be_uninstalled():
			os.unlink(self.get_installed_path())

	def installed(self):
		""" Return true if the package is installed, false otherwise."""
		return os.path.exists(self.get_installed_path())

################################################################################

class File:
	def __init__(self, package, name, md5, crc32, size):
		self.package = package
		self.name = name
		self.md5 = md5
		self.crc32 = crc32
		self.size = size

	def get_pool_path(self):
		""" Return the physical path to the file in the pool."""
		md5 = binascii.hexlify(self.md5)
		return os.path.join(pool_dir, md5[:2], md5[2:]) + '.gz'

	def available(self):
		""" Return true iff the file is available locally."""
		return os.path.exists(self.get_pool_path())

################################################################################

import unittest, shutil, time
from downloader import MockDownloader

class TestRapid(unittest.TestCase):
	test_dir = os.path.realpath('.test-rapid')

	def setUp(self):
		set_spring_dir(self.test_dir)

		# Speed up the test because if pool is present the 256 pool
		# directories are created on demand instead of beforehand.
		mkdir_p(pool_dir)

		self.rapid = Rapid()

		if True:   # False to use real Downloader, True to use MockDownloader
			self.rapid.downloader = MockDownloader()
			www = self.rapid.downloader.www
			www[master_url] = gzip_string(',http://ts1,,\n')
			www['http://ts1/versions.gz'] = gzip_string('xta:latest,1234,dependency,XTA 9.6\n,5678,,dependency\n')
			www['http://ts1/packages/1234.sdp'] = gzip_string('\3foo' + binascii.unhexlify('d41d8cd98f00b204e9800998ecf8427e') + 8 * '\0')
			www['http://ts1/packages/5678.sdp'] = gzip_string('')
			www['http://ts1/streamer.cgi?1234'] = struct.pack('>L', len(gzip_string(''))) + gzip_string('')

	def tearDown(self):
		shutil.rmtree(self.test_dir)

	def test_get_repositories(self):
		self.rapid.get_repositories()

	def test_get_packages_by_name(self):
		self.rapid.get_packages_by_name()

	def test_get_package_by_name(self):
		self.assertFalse(self.rapid.get_package_by_name('XXXXXX'))
		self.assertTrue(self.rapid.get_package_by_name('XTA 9.6'))

	def test_get_packages_by_tag(self):
		self.rapid.get_packages_by_tag()

	def test_get_package_by_tag(self):
		self.assertFalse(self.rapid.get_package_by_tag('XXXXXX'))
		self.assertTrue(self.rapid.get_package_by_tag('xta:latest'))

	def test_get_packages(self):
		self.rapid.get_packages()

	def test_get_installed_packages(self):
		self.rapid.get_installed_packages()

	def test_get_not_installed_packages(self):
		self.rapid.get_not_installed_packages()

	def install(self, p):
		for d in p.dependencies:
			self.install(d)
		p.install()

	def test_install_uninstall(self):
		p = self.rapid.get_package_by_tag('xta:latest')
		self.install(p)
		self.assertFalse(p.get_missing_files())
		self.assertTrue(os.path.exists(p.get_files()[0].get_pool_path()))
		self.assertTrue(os.path.exists(os.path.join(package_dir, '1234.sdp')))
		p.uninstall()
		self.assertFalse(os.path.exists(os.path.join(package_dir, '1234.sdp')))

	def test_install_missing_dependency(self):
		p = self.rapid.get_package_by_tag('xta:latest')
		p.install()
		self.assertFalse(p.installed())   # install should have failed

	def test_uninstall_dependency_check(self):
		p = self.rapid.get_package_by_tag('xta:latest')
		self.install(p)
		d = self.rapid.get_package_by_name('dependency')
		d.uninstall()
		self.assertTrue(d.installed())   # uninstall should have failed

if __name__ == '__main__':
	unittest.main()
